/*
	This is a Virtual Window Manager for use under Win32 Shells that
	support the LiteStep module standard.

	Copyright (C) 2001 Chris Rempel

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <tchar.h>
#include <windows.h>
#include <stdlib.h>


#define VW_NEXT(x) (x < nDesks-1 ? x+1 : 0)
#define VW_PREV(x) (x > 0 ? x-1 : nDesks-1)

__declspec( dllimport ) BOOL AddBangCommand( LPCTSTR, VOID (__cdecl *)(HWND, LPCTSTR) );
__declspec( dllimport ) BOOL RemoveBangCommand( LPCTSTR );

extern VOID switchToVW( UINT );
extern VOID gatherToVW( UINT, char );
extern VOID setNumDesks( UINT );

extern UINT nCurVW;
extern UINT nDesks;
extern BOOL bVertical;



INT getVW( LPCTSTR args )
{
	INT retVal = -1;

	if (args != NULL && *args != 0)
	{

		if (!_tcsicmp(_T(".home"), args))
			retVal = 0;

		else if (!_tcsicmp(_T(".end"), args))
			retVal = nDesks-1;

		else if (!_tcsicmp(_T(".next"), args))
			retVal = VW_NEXT(nCurVW);

		else if (!_tcsicmp(_T(".prev"), args))
			retVal = VW_PREV(nCurVW);

		else
			retVal = _ttoi(args);

	}

	if (retVal < 0 || (UINT)retVal >= nDesks)
		retVal = nCurVW;

	return retVal;
}


/* Bang Commands */
VOID bVWMSwitchTo( HWND caller, LPCTSTR args )
{
	switchToVW( getVW( args ) );

	return;

		UNREFERENCED_PARAMETER( caller );
}

VOID bVWMNext( HWND caller, LPCTSTR args )
{
	switchToVW( VW_NEXT(nCurVW) );

	return;

		UNREFERENCED_PARAMETER( caller );
		UNREFERENCED_PARAMETER( args );
}

VOID bVWMPrev( HWND caller, LPCTSTR args )
{
	switchToVW( VW_PREV(nCurVW) );

	return;

		UNREFERENCED_PARAMETER( caller );
		UNREFERENCED_PARAMETER( args );
}

VOID bVWMGather( HWND caller, LPCTSTR args )
{
	gatherToVW( getVW( args ), 0x03 );

	return;

		UNREFERENCED_PARAMETER( caller );
}

VOID bVWMGatherH( HWND caller, LPCTSTR args )
{
	gatherToVW( getVW( args ), 0x01 );

	return;

		UNREFERENCED_PARAMETER( caller );
}

VOID bVWMGatherV( HWND caller, LPCTSTR args )
{
	gatherToVW( getVW( args ), 0x02 );

	return;

		UNREFERENCED_PARAMETER( caller );
}

VOID bVWMDesks( HWND caller, LPCTSTR args )
{

	if (args != NULL && isdigit( *args ))
	{
		UINT tmp = _ttoi( args );

		if (tmp != nDesks)
		{
			setNumDesks(tmp);
			// Call gather command for windows outside of new space
		}
	}

	return;

		UNREFERENCED_PARAMETER( caller );
}

#if 0
VOID bVWMShiftOrigin( HWND caller, LPCTSTR args )
{

args=args;

	return;

		UNREFERENCED_PARAMETER( caller );
}

VOID bVWMMoveWindowTo( HWND caller, LPCTSTR args )
{

args=args;

	return;

		UNREFERENCED_PARAMETER( caller );
}
#endif


	/*
	These should be in the "utility.c" file, but I didn't want to
	extern all of the bang command functions.
	*/
VOID RegisterBangCmds( VOID )
{

	AddBangCommand(_T("!BVWMSWITCHTO"), bVWMSwitchTo);
	AddBangCommand(_T("!BVWMNEXT"), bVWMNext);
	AddBangCommand(_T("!BVWMPREV"), bVWMPrev);
	AddBangCommand(_T("!BVWMGATHER"), bVWMGather);
	AddBangCommand(_T("!BVWMGATHERH"), bVWMGatherH);
	AddBangCommand(_T("!BVWMGATHERV"), bVWMGatherV);
	AddBangCommand(_T("!BVWMDESKS"), bVWMDesks);
#if 0
	AddBangCommand(_T("!BVWMSHIFTORIGIN"), bVWMShiftOrigin);
	AddBangCommand(_T("!BVWMMOVEWINDOWTO"), bVWMMoveWindowTo);
#endif
	return;
}

VOID UnRegisterBangCmds( VOID )
{

	RemoveBangCommand(_T("!BVWMSWITCHTO"));
	RemoveBangCommand(_T("!BVWMNEXT"));
	RemoveBangCommand(_T("!BVWMPREV"));
	RemoveBangCommand(_T("!BVWMGATHER"));
	RemoveBangCommand(_T("!BVWMGATHERH"));
	RemoveBangCommand(_T("!BVWMGATHERV"));
	RemoveBangCommand(_T("!BVWMDESKS"));
#if 0
	RemoveBangCommand(_T("!BVWMSHIFTORIGIN"));
	RemoveBangCommand(_T("!BVWMMOVEWINDOWTO"));
#endif
	return;
}

