/*
	This is a Virtual Window Manager for use under Win32 Shells that
	support the LiteStep module standard.

	Copyright (C) 2001 Chris Rempel

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <tchar.h>
#include <windows.h>
#include <stdlib.h>


/* Utility defines */
#define SHELL_WINDOW_ID 0x49474541
#define VW_SPACING 10
#define VW_WIDTH (nDTWidth + VW_SPACING)
#define VW_HEIGHT (nDTHeight + VW_SPACING)
#define VW_ISVALID(x) ((x >= 0) && (x < nDesks))

/* LSAPI.DLL FUNCTION PROTOTYPES */
__declspec( dllimport ) INT GetRCInt( LPCTSTR, INT );
__declspec( dllimport ) BOOL GetRCBool( LPCTSTR, BOOL );
__declspec( dllimport ) BOOL GetRCLine( LPCTSTR, LPTSTR, size_t, LPCTSTR );
__declspec( dllimport ) BOOL LCReadNextConfig( UINT, LPCTSTR, LPTSTR, DWORD );
__declspec( dllimport ) INT LCTokenize( LPCTSTR, LPTSTR*, DWORD, LPTSTR );
__declspec( dllimport ) HINSTANCE LSExecute( HWND, LPCTSTR, INT );
__declspec( dllimport ) BOOL match( char *, char * );

/* Our globals */
HWND hTopWnd;
UINT nCurVW, nDesks;
INT nDTWidth, nDTHeight;
INT nWAWidth, nWAHeight;
BOOL bVertical;
BOOL bStrictGather;
LPTSTR quitCmd;
LPTSTR *cmdList;
LPTSTR *stickyList;
UINT numSticky;

VOID freeCommands( VOID )
{
	UINT i;

	if (cmdList != NULL){
		for (i=0;i<nDesks;i++)
		{
			if (cmdList[i] != NULL)
			{
				free( cmdList[i] );
				cmdList[i] = NULL;
			}
		}

		free( cmdList );
		cmdList = NULL;
	}

	return;
}

VOID readCommands( VOID )
{

	LPTSTR tokens[2];
	_TCHAR buffer[4096];
	_TCHAR token[128];
	_TCHAR cmd[4096];

	tokens[0] = token;
	tokens[1] = token;

	while (LCReadNextConfig(1, _T("*bVWMExec"), buffer, sizeof(buffer)/sizeof(_TCHAR)))
	{
		if (cmdList == NULL)
		{
			cmdList = (LPTSTR*)calloc(nDesks, sizeof(LPTSTR));
		}

		if (cmdList != NULL)
		{
			UINT count;

			*cmd = 0;
			count = LCTokenize(buffer, tokens, 2, cmd);
			if ((count >= 2) && (*cmd != 0))
			{
				count = _ttoi(tokens[1]);
				if ((count >= 0) && (count < nDesks) && (cmdList[count] == NULL))
				{
					cmdList[count] = (LPTSTR)malloc((_tcslen(cmd)+1)*sizeof(_TCHAR));

					if (cmdList[count] != NULL)
					{
						_tcscpy(cmdList[count], cmd);
					}
				}
			}
		}
	}

	return;
}

VOID freeSticky( VOID )
{
	UINT i;

	if (stickyList != NULL)
	{
		for (i=0;i<numSticky;i++)
		{
			if (stickyList[i] != NULL)
			{
				free( stickyList[i] );
				stickyList[i] = NULL;
			}
		}
		numSticky = 0;

		free( stickyList );
		stickyList = NULL;
	}

	return;
}

VOID readSticky( VOID )
{

	LPTSTR* tmpList;
	_TCHAR buffer[1024];
	_TCHAR cmd[1024];

	LPTSTR tokens[2];
	tokens[0] = cmd;
	tokens[1] = cmd;

	while (LCReadNextConfig(1, _T("*bVWMSticky"), buffer, sizeof(buffer)/sizeof(_TCHAR)))
	{
		*cmd = 0;
		if ((LCTokenize( buffer, tokens, 2, NULL ) >= 2) && (*cmd != 0))
		{
			tmpList = stickyList;

			if (stickyList == NULL)
			{
				numSticky = 0;
				stickyList = (LPTSTR*)malloc(sizeof(LPTSTR));
			}
			else
			{
				stickyList = (LPTSTR*)realloc(stickyList, (numSticky+1)*sizeof(LPTSTR));

				if (stickyList == NULL)
				{
					stickyList = tmpList;
				}
				else
				{
					tmpList = NULL;
				}
			}

			if (tmpList == NULL && stickyList != NULL)
			{
				stickyList[numSticky] = (LPTSTR)malloc((_tcslen(cmd)+1)*sizeof(_TCHAR));

				if (stickyList[numSticky] != NULL)
				{
					_tcscpy(stickyList[numSticky++], cmd);
				}
			}
		}
	}

	return;
}

VOID setNumDesks( UINT nNew )
{
	if (nNew < 1)
		nNew = 1;

	nDesks = nNew;

	freeCommands();
	readCommands();

	// Call gather command for windows outside of new space

	return;
}


VOID LoadSettings( VOID )
{
	RECT r;
	_TCHAR tmp[4096];

	GetClientRect(GetDesktopWindow(),&r);
	nDTWidth = r.right;
	nDTHeight = r.bottom;

	SystemParametersInfo(SPI_GETWORKAREA,0,&r,0);
	nWAWidth = r.right;
	nWAHeight = r.bottom;

	bVertical = GetRCBool( _T("bVWMVertical"), TRUE );
	bStrictGather = GetRCBool( _T("bVWMStrictGather"), TRUE );

	setNumDesks(GetRCInt( _T("bVWMDesks"), 2 ));

	//readCommands(); <- done in setNumDesks()
	readSticky();

	*tmp = 0;
	GetRCLine( _T("bVWMQuitCmd"), tmp, 4096, _T("") );
	if (*tmp != 0)
	{
		quitCmd = (LPTSTR)malloc((_tcslen(tmp)+1)*sizeof(_TCHAR));
		if (quitCmd != NULL)
			_tcscpy(quitCmd, tmp);
	}

	return;
}


VOID FreeSettings( VOID )
{

	if (quitCmd != NULL)
	{
		LSExecute( NULL, quitCmd, 0 );
		free( quitCmd );
		quitCmd = NULL;
	}

	freeSticky();
	freeCommands();

	return;
}


VOID executeCmd( UINT vw )
{
	if (VW_ISVALID(vw) && cmdList != NULL && cmdList[vw] != NULL)
		LSExecute( NULL, cmdList[vw], 0 );

	return;
}


BOOL CALLBACK EnumWindowsProc( HWND hwnd, LPARAM lParam )
{
	if (!(GetWindowLong(hwnd, GWL_EXSTYLE) & WS_EX_TOPMOST))
	{
		hTopWnd = hwnd;
		return FALSE;
	}
	return TRUE;
		UNREFERENCED_PARAMETER( lParam );
}


BOOL isSticky( HWND hwnd )
{
	UINT i;
	_TCHAR text[256];
	BOOL retVal;

	text[255] = 0;
	retVal = FALSE;
	for (i=0;retVal == FALSE && i<numSticky;i++)
	{
		GetClassName(hwnd, text, 255);
		if (!match(stickyList[i], text))
		{
			GetWindowText(hwnd, text, 255);
			if (match(stickyList[i], text))
				retVal = TRUE;
		}
		else
			retVal = TRUE;
	}

	return retVal;
}


VOID switchToVW( UINT vw )
{
	if ( vw != nCurVW && VW_ISVALID(vw) ){
		INT add;
		HWND hwnd;
		HDWP dwp;
		RECT r;
		BOOL next;

		add = ((vw % nDesks) - (nCurVW % nDesks)) * (bVertical ? VW_HEIGHT:VW_WIDTH);

		next = FALSE;
		EnumWindows(EnumWindowsProc, 0);
		hwnd = GetWindow(hTopWnd, GW_HWNDLAST);

		dwp = BeginDeferWindowPos(1);

		do
		{

			if (next)
				hwnd = GetWindow(hwnd, GW_HWNDPREV);

			next = hwnd != NULL;

			if (next)
			{
				if (!(GetWindowLong(hwnd, GWL_STYLE) & WS_VISIBLE))
					continue;
				if (GetWindowLong(hwnd, GWL_USERDATA) == SHELL_WINDOW_ID)
					continue;
				if (isSticky(hwnd))
					continue;

				GetWindowRect(hwnd, &r);
				if (bVertical)
				{
					r.top-=add;
					r.bottom-=add;
				}
				else
				{
					r.left-=add;
					r.right-=add;
				}
				dwp = DeferWindowPos(dwp, hwnd, NULL, r.left, r.top, r.right-r.left, r.bottom-r.top, SWP_NOZORDER|SWP_NOACTIVATE);

			}

		}while (next);

		nCurVW = vw;

		EndDeferWindowPos(dwp);

		executeCmd( vw );
	}

	return;
}


VOID updateRectH( RECT *r )
{
	INT size;

	size = abs(r->right - r->left);

		/* Initial x/width position calculation */
	if (r->left > nDTWidth)
	{
		r->left %= VW_WIDTH;
		if (r->left > nDTWidth)
			r->left -= VW_WIDTH;
		r->right = r->left + size;
	}
	else if (r->left < -(VW_SPACING))
	{
		r->left %= VW_WIDTH;
		if (r->left < -(VW_SPACING))
			r->left += VW_WIDTH;;
		r->right = r->left + size;
	}

		/* Don't allow a window to be in more then one VW at a time if possible. */
	if (!bStrictGather && r->right > VW_WIDTH)
	{
		if (r->left > r->right-nDTWidth)
			r->left -= (r->right-nDTWidth);
		else
		{
			if (size - VW_WIDTH < VW_SPACING)
				r->left = (size-nDTWidth)/-2;
			else
				r->left = 0;
		}
		r->right = r->left + size;
	}

	return;
}


VOID updateRectV( RECT *r )
{
	INT size;

	size = abs(r->bottom - r->top);

		/* Initial x/width position calculation */
	if (r->top > nDTHeight)
	{
		r->top %= VW_HEIGHT;
		if (r->top > nDTHeight)
			r->top -= VW_HEIGHT;
		r->bottom = r->top + size;
	}
	else if (r->top < -(VW_SPACING))
	{
		r->top %= VW_HEIGHT;
		while (r->top < -(VW_SPACING))
			r->top += VW_HEIGHT;;
		r->bottom = r->top + size;
	}

		/* Don't allow a window to be in more then one VW at a time if possible. */
	if (!bStrictGather && r->bottom > VW_HEIGHT)
	{
		if (r->top > r->bottom-nDTHeight)
			r->top -= (r->bottom-nDTHeight);
		else
		{
			if (size - VW_HEIGHT < VW_SPACING)
				r->top = (size-nDTHeight)/-2;
			else
				r->top = 0;
		}
		r->bottom = r->top + size;
	}

	return;
}


VOID gatherToVW( UINT vw, char type )
{
	if ( VW_ISVALID(vw) ){
		HWND hwnd;
		HDWP dwp;
		RECT r;
		BOOL next;

		next = FALSE;
		EnumWindows(EnumWindowsProc, 0);
		hwnd = GetWindow(hTopWnd, GW_HWNDLAST);
		dwp = BeginDeferWindowPos(1);

		do
		{

			if (next)
				hwnd = GetWindow(hwnd, GW_HWNDPREV);

			next = hwnd != NULL;

			if (next)
			{
				if (!(GetWindowLong(hwnd, GWL_STYLE) & WS_VISIBLE))
					continue;
				if ((GetWindowLong(hwnd, GWL_STYLE) & WS_MINIMIZE))
					continue;
				if (GetWindowLong(hwnd, GWL_USERDATA) == SHELL_WINDOW_ID)
					continue;

				GetWindowRect(hwnd, &r);

				if (type & 0x01)
					updateRectH( &r );
				if (type & 0x02)
					updateRectV( &r );

				dwp = DeferWindowPos(dwp, hwnd, NULL, r.left, r.top, r.right-r.left, r.bottom-r.top, SWP_NOZORDER|SWP_NOACTIVATE);
			}

		}while (next);

		EndDeferWindowPos(dwp);

		if (nCurVW != vw)
		{
			nCurVW = vw;
			executeCmd( vw );
		}
	}

	return;
}

#if 0
UINT getVWFromHwnd( HWND hwnd )
{
	UINT vw;
	RECT r;

	GetWindowRect(hwnd, &r);


	vw = (vw % nDesks) - (nCurVW % nDesks);

	//add = ((vw % nDesks) - (nCurVW % nDesks)) * (bVertical ? VW_HEIGHT:VW_WIDTH);

	return vw;
}
#endif
